==> srtp-1.0.1.tgz: srtp-1.0.1/crypto/replay/rdbx.c <==
/*
 * replay-db-ext.c
 *
 * a replay database with extended range, using a rollover counter
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */

/*
 *	
 * Copyright  (c) 2001, 2002, Cisco Systems, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 *   Neither the name of the Cisco Systems, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "rdbx.h"

#define rdbx_high_bit_in_bitmask 127

/*
 * from draft-ietf-avt-srtp-00.txt:
 *
 * A receiver reconstructs the index i of a packet with sequence
 *  number s using the estimate
 *
 * i = 65,536 * t + s,
 *
 * where t is chosen from the set { r-1, r, r+1 } such that i is
 * closest to the value 65,536 * r + s_l.  If the value r+1 is used,
 * then the rollover counter r in the cryptographic context is
 * incremented by one (if the packet containing s is authentic).
 */



/*
 * rdbx implementation notes
 *
 * A xtd_seq_num_t is essentially a sequence number for which some of
 * the data on the wire are implicit.  It consists of a
 * rollover_counter and a sequence_number; the sequence number is the
 * explicit part, and the rollover_counter is the implicit part.  
 *
 * Upon receiving a sequence_number (e.g. in a newly received SRTP
 * packet), the complete xtd_seq_num_t can be estimated by using a
 * local xtd_seq_num_t as a basis.  This is done using the function
 * index_guess(&local, &guess, seq_from_packet).  This function
 * returns the difference of the guess and the local value.  The local
 * xtd_seq_num_t can be moved forward to the guess using the function
 * index_advance(&guess, delta), where delta is the difference.
 * 
 *
 * A rdbx_t consists of a xtd_seq_num_t and a bitmask.  The index is highest
 * sequence number that has been received, and the bitmask indicates
 * which of the recent indicies have been received as well.  The
 * highest bit in the bitmask corresponds to the index in the bitmask.
 */


void
index_init(xtd_seq_num_t *pi) {
  pi->roc = 0;
  pi->seq = 0;
}


void
index_advance(xtd_seq_num_t *pi, sequence_number_t s) {
  rollover_counter_t tmp;

  tmp = (rollover_counter_t) s + pi->seq;
  pi->seq = (sequence_number_t) tmp;
  pi->roc += (tmp >> (4*sizeof(rollover_counter_t))); 
}

void
index_add(xtd_seq_num_t *pi, sequence_number_t s, rollover_counter_t r) {
  rollover_counter_t tmp;

  tmp = (rollover_counter_t) s + pi->seq;
  pi->seq = (sequence_number_t) tmp;
  pi->roc += (tmp >> (4*sizeof(rollover_counter_t)));
  pi->roc += r;
}


/*
 * index_guess(local, guess, s)
 * 
 * given a xtd_seq_num_t local (which represents the last
 * known-to-be-good received xtd_seq_num_t) and a sequence number s
 * (from a newly arrived packet), sets the contents of *guess to
 * contain the best guess of the packet index to which s corresponds,
 * and returns the difference between *guess and *local
 *
 * nota bene - the output is a signed integer, don't cast to a uint
 */

int
index_guess(const xtd_seq_num_t *local,
		   xtd_seq_num_t *guess,
		   sequence_number_t s) {
  int difference;
  
  if (local->seq < seq_num_median) {
    if (s - local->seq > seq_num_median) {
      guess->roc = local->roc - 1;
      difference = seq_num_max - s + local->seq;
    } else {
      guess->roc = local->roc;
      difference = s - local->seq;
    }
  } else {
    if (local->seq - seq_num_median > s) {
      guess->roc = local->roc+1;
      difference = seq_num_max - local->seq + s;
    } else {
      difference = s - local->seq;
      guess->roc = local->roc;
    }
  }
  guess->seq = s;
  
  return difference;
}

/*
 * rdbx
 *
 */


/*
 *  rdbx_init(&r) initalizes the rdbx_t pointed to by r 
 */

replay_status
rdbx_init(rdbx_t *rdbx) {
  v128_set_to_zero(&rdbx->bitmask);
  index_init(&rdbx->index);

  return replay_status_ok;
}


/*
 * rdbx_check(&r, delta) checks to see if the xtd_seq_num_t
 * which is at rdbx->index + delta is in the rdb
 */

inline replay_status
rdbx_check(const rdbx_t *rdbx, int delta) {
  
  if (delta > 0) {       /* if delta is positive, it's good */
    return replay_check_ok;
  } else if (rdbx_high_bit_in_bitmask + delta < 0) {   
                         /* if delta is lower than the bitmask, it's bad */
    return replay_check_fail; 
  } else if (v128_get_bit(&rdbx->bitmask, 
			  rdbx_high_bit_in_bitmask + delta) == 1) {
                         /* delta is within the window, so check the bitmask */
    return replay_check_fail;    
  } else {               /* otherwise, the index is okay */
  }

  return replay_check_ok;   /* keep compiler silent */
}

/*
 * rdbx_add_index adds the xtd_seq_num_t at rdbx->window_start + d to
 * replay_db (and does *not* check if that xtd_seq_num_t appears in db)
 *
 * this function should be called only after replay_check has
 * indicated that the index does not appear in the rdbx, e.g., a mutex
 * should protect the rdbx between these calls
 */

inline replay_status
rdbx_add_index(rdbx_t *rdbx, int delta) {
  
  if (delta > 0) {
    /* shift forward by delta */
    index_advance(&rdbx->index, delta);
    v128_left_shift(&rdbx->bitmask, delta);
    v128_set_bit(&rdbx->bitmask, 127);
  } else {
    /* delta is in window, so flip bit in bitmask */
    v128_set_bit(&rdbx->bitmask, -delta);
  }

  /* note that we need not consider the case that delta == 0 */
  
  
  return replay_status_ok;
}



/*
 * rdbx_estimate_index(rdbx, guess, s)
 * 
 * given an rdbx and a sequence number s (from a newly arrived packet),
 * sets the contents of *guess to contain the best guess of the packet
 * index to which s corresponds, and returns the difference between
 * *guess and the locally stored synch info
 */

inline int
rdbx_estimate_index(const rdbx_t *rdbx,
		    xtd_seq_num_t *guess,
		    sequence_number_t s) {

  /*
   * if the sequence number and rollover counter in the rdbx are
   * non-zero, then use the index_guess(...) function, otherwise, just
   * set the rollover counter to zero (since the index_guess(...)
   * function might incorrectly guess that the rollover counter is
   * 0xffffffff)
   */

  if (rdbx->index.seq && rdbx->index.roc) 
    return index_guess(&rdbx->index, guess, s);
  
  guess->roc = 0;
  guess->seq = s;
  return s;
}

==> srtp-1.0.2.tgz: srtp-1.0.2/crypto/replay/rdbx.c <==
/*
 * replay-db-ext.c
 *
 * a replay database with extended range, using a rollover counter
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */

/*
 *	
 * Copyright  (c) 2001, 2002, Cisco Systems, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 *   Neither the name of the Cisco Systems, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "rdbx.h"

#define rdbx_high_bit_in_bitmask 127

/*
 * from draft-ietf-avt-srtp-00.txt:
 *
 * A receiver reconstructs the index i of a packet with sequence
 *  number s using the estimate
 *
 * i = 65,536 * t + s,
 *
 * where t is chosen from the set { r-1, r, r+1 } such that i is
 * closest to the value 65,536 * r + s_l.  If the value r+1 is used,
 * then the rollover counter r in the cryptographic context is
 * incremented by one (if the packet containing s is authentic).
 */



/*
 * rdbx implementation notes
 *
 * A xtd_seq_num_t is essentially a sequence number for which some of
 * the data on the wire are implicit.  It consists of a
 * rollover_counter and a sequence_number; the sequence number is the
 * explicit part, and the rollover_counter is the implicit part.  
 *
 * Upon receiving a sequence_number (e.g. in a newly received SRTP
 * packet), the complete xtd_seq_num_t can be estimated by using a
 * local xtd_seq_num_t as a basis.  This is done using the function
 * index_guess(&local, &guess, seq_from_packet).  This function
 * returns the difference of the guess and the local value.  The local
 * xtd_seq_num_t can be moved forward to the guess using the function
 * index_advance(&guess, delta), where delta is the difference.
 * 
 *
 * A rdbx_t consists of a xtd_seq_num_t and a bitmask.  The index is highest
 * sequence number that has been received, and the bitmask indicates
 * which of the recent indicies have been received as well.  The
 * highest bit in the bitmask corresponds to the index in the bitmask.
 */


void
index_init(xtd_seq_num_t *pi) {
  pi->roc = 0;
  pi->seq = 0;
}


void
index_advance(xtd_seq_num_t *pi, sequence_number_t s) {
  rollover_counter_t tmp;

  tmp = (rollover_counter_t) s + pi->seq;
  pi->seq = (sequence_number_t) tmp;
  pi->roc += (tmp >> (4*sizeof(rollover_counter_t))); 
}

void
index_add(xtd_seq_num_t *pi, sequence_number_t s, rollover_counter_t r) {
  rollover_counter_t tmp;

  tmp = (rollover_counter_t) s + pi->seq;
  pi->seq = (sequence_number_t) tmp;
  pi->roc += (tmp >> (4*sizeof(rollover_counter_t)));
  pi->roc += r;
}


/*
 * index_guess(local, guess, s)
 * 
 * given a xtd_seq_num_t local (which represents the last
 * known-to-be-good received xtd_seq_num_t) and a sequence number s
 * (from a newly arrived packet), sets the contents of *guess to
 * contain the best guess of the packet index to which s corresponds,
 * and returns the difference between *guess and *local
 *
 * nota bene - the output is a signed integer, don't cast to a uint
 */

int
index_guess(const xtd_seq_num_t *local,
		   xtd_seq_num_t *guess,
		   sequence_number_t s) {
  int difference;
  
  if (local->seq < seq_num_median) {
    if (s - local->seq > seq_num_median) {
      guess->roc = local->roc - 1;
      difference = seq_num_max - s + local->seq;
    } else {
      guess->roc = local->roc;
      difference = s - local->seq;
    }
  } else {
    if (local->seq - seq_num_median > s) {
      guess->roc = local->roc+1;
      difference = seq_num_max - local->seq + s;
    } else {
      difference = s - local->seq;
      guess->roc = local->roc;
    }
  }
  guess->seq = s;
  
  return difference;
}

/*
 * rdbx
 *
 */


/*
 *  rdbx_init(&r) initalizes the rdbx_t pointed to by r 
 */

replay_status
rdbx_init(rdbx_t *rdbx) {
  v128_set_to_zero(&rdbx->bitmask);
  index_init(&rdbx->index);

  return replay_status_ok;
}


/*
 * rdbx_check(&r, delta) checks to see if the xtd_seq_num_t
 * which is at rdbx->index + delta is in the rdb
 */

inline replay_status
rdbx_check(const rdbx_t *rdbx, int delta) {
  
  if (delta > 0) {       /* if delta is positive, it's good */
    return replay_check_ok;
  } else if (rdbx_high_bit_in_bitmask + delta < 0) {   
                         /* if delta is lower than the bitmask, it's bad */
    return replay_check_fail; 
  } else if (v128_get_bit(&rdbx->bitmask, 
			  rdbx_high_bit_in_bitmask + delta) == 1) {
                         /* delta is within the window, so check the bitmask */
    return replay_check_fail;    
  } else {               /* otherwise, the index is okay */
  }

  return replay_check_ok;   /* keep compiler silent */
}

/*
 * rdbx_add_index adds the xtd_seq_num_t at rdbx->window_start + d to
 * replay_db (and does *not* check if that xtd_seq_num_t appears in db)
 *
 * this function should be called only after replay_check has
 * indicated that the index does not appear in the rdbx, e.g., a mutex
 * should protect the rdbx between these calls
 */

inline replay_status
rdbx_add_index(rdbx_t *rdbx, int delta) {
  
  if (delta > 0) {
    /* shift forward by delta */
    index_advance(&rdbx->index, delta);
    v128_left_shift(&rdbx->bitmask, delta);
    v128_set_bit(&rdbx->bitmask, 127);
  } else {
    /* delta is in window, so flip bit in bitmask */
    v128_set_bit(&rdbx->bitmask, -delta);
  }

  /* note that we need not consider the case that delta == 0 */
  
  
  return replay_status_ok;
}



/*
 * rdbx_estimate_index(rdbx, guess, s)
 * 
 * given an rdbx and a sequence number s (from a newly arrived packet),
 * sets the contents of *guess to contain the best guess of the packet
 * index to which s corresponds, and returns the difference between
 * *guess and the locally stored synch info
 */

inline int
rdbx_estimate_index(const rdbx_t *rdbx,
		    xtd_seq_num_t *guess,
		    sequence_number_t s) {

  /*
   * if the sequence number and rollover counter in the rdbx are
   * non-zero, then use the index_guess(...) function, otherwise, just
   * set the rollover counter to zero (since the index_guess(...)
   * function might incorrectly guess that the rollover counter is
   * 0xffffffff)
   */

  if (rdbx->index.roc || (rdbx->index.seq > seq_num_median)) 
    return index_guess(&rdbx->index, guess, s);
  
  guess->roc = 0;
  guess->seq = s;
  return s;
}
